header "Reasoning about PreFormulas"

theory PreFormulas
imports Semantics
begin

constdefs
   PreValid :: "('w::world) form \<Rightarrow> bool"
   [simp]: "PreValid A \<equiv> \<forall> w. w \<Turnstile> A"

syntax
  "_PreValid"      :: "lift \<Rightarrow> bool"     ("(|~ _)" 5)

translations
  "_PreValid"  \<rightleftharpoons> "PreValid"

lemma prefD[dest]: "|~ A \<Longrightarrow> w \<Turnstile> A"
proof -
  assume a:"|~ A"
  from a have "ALL w. w \<Turnstile> A" by (auto simp add: PreValid_def)
  thus ?thesis ..
qed

lemma prefI[simp]: assumes P1:"(\<And> w. w \<Turnstile> A)" shows "|~ A"
proof -
  fix w
  from P1 have "w \<Turnstile> A" by blast
  thus ?thesis by auto
qed

lemma prefeq_reflection[simp]: assumes P1: "|~ x=y" shows  "(x \<equiv> y)"
proof -
  from P1 have "ALL w. w |= x=y" by (unfold PreValid_def)
  hence P2: "ALL w. x w = y w" by (unfold unl_lift2)
  hence P3:"x=y"
  proof -
    from P2 have "\<And> w. x w = y w" ..
    thus ?thesis by (rule ext)
  qed
  thus "x \<equiv> y" by (rule "eq_reflection")
qed

method_setup pref_unlift = 
{*
let 
  fun pref_unlift_tac args ctxt = 
      let val intensional_rews = ProofContext.get_thms ctxt (PureThy.Name "intensional_rews")
          val prefI = ProofContext.get_thm ctxt (PureThy.Name "prefI")
      in 
        (Method.METHOD (fn chained_facts => 
           (fn th => 
           Seq.single (rewrite_rule intensional_rews ((prefI RS th) handle _ => th)))))
      end
in 
  pref_unlift_tac 
end
*} "pref_unlift tactic lifted to being a method."

lemma pref_eq: "|~ X = Y \<Longrightarrow> X = Y"
  by auto

lemma pref_iffI: assumes h1: "|~ F \<longrightarrow> G"
                 and   h2: "|~ G \<longrightarrow> F"
                shows      "|~ F = G"
proof (pref_unlift,rule iffI)
  fix w
  assume "w \<Turnstile> F"
  with h1[THEN prefD] show "w \<Turnstile> G" by auto
next
  fix w
  assume "w \<Turnstile> G"
  with h2[THEN prefD] show "w \<Turnstile> F" by auto
qed

lemma pref_iffD1: assumes h: "|~ F = G"
  shows "|~ F \<longrightarrow> G"
  using h[THEN prefD] by auto

lemma pref_iffD2: assumes h: "|~ F = G"
  shows "|~ G \<longrightarrow> F"
  using h[THEN prefD] by auto

lemma unl_pref_imp: assumes h: "|~ F \<longrightarrow> G"
       shows "\<And> w. w \<Turnstile> F \<Longrightarrow> w \<Turnstile> G"
  using h[THEN prefD]
  by auto

section "Unchanged formulas"

text {* Many of the TLA* axioms only requires
 a state function witness which leaves the state
 space unchanged. An obvious witness is the @{term
id} function. The lemmas require that the given 
 formula is invariant under stuttering.
*}

lemma pre_id_unch: assumes h: "stutinv F"
  shows "|~ F \<and> Unchanged id \<longrightarrow> \<circ>F"
proof (pref_unlift,rule impI,erule conjE)
  fix s
  assume a1: "s \<Turnstile> F"
    and  a2: "s \<Turnstile> Unchanged id"
  from h have "\<forall> \<sigma> \<tau>. \<sigma> \<approx> \<tau> \<longrightarrow> (\<sigma> \<Turnstile> F) = (\<tau> \<Turnstile> F)"
    by (simp add: stutinv_def)
  hence g1: "s \<approx> (tail s) \<longrightarrow> (s \<Turnstile> F) = ((tail s) \<Turnstile> F)"
    by blast
  thm unch_defs
  from a2 have "(id (second s) = id (first s))"
    by (simp add: unch_defs)
  hence "first s = second s" by auto
  hence "s \<approx> (tail s)" by (rule addfirststut)
  with g1 a1 have "(tail s) \<Turnstile> F" by simp
  thus "s \<Turnstile> \<circ>F" by (unfold nexts_def)
qed

lemma pre_ex_unch: assumes h: "stutinv F"
   shows "\<exists> (v :: 'a \<Rightarrow> 'a). ( |~ F \<and> Unchanged v \<longrightarrow> \<circ>F)" 
proof
  from h show "|~ F \<and> Unchanged id \<longrightarrow> \<circ>F" by (rule pre_id_unch)
qed

lemma unch1: "|~ Unchanged (x,y) \<longrightarrow> $x = x$ \<and> $y = y$"
  apply (pref_unlift)
  apply (unfold unch_def before_def after_def nexts_def)
  apply auto
  done

lemma unch2: assumes h: "|~ Unchanged (x,y)"
       shows  "|~ $x = x$ \<and> $y = y$"
  using h
  apply (unfold unch_def before_def after_def nexts_def)
  apply (rule prefI)
  apply auto   
  done

lemma unch3: assumes h: "|~ $x = x$ \<and> $y = y$"
  shows "|~ Unchanged (x,y)"
  using h
  apply (unfold unch_def before_def after_def nexts_def)
  apply (rule prefI)
  apply auto   
  done

lemma unch4: "|~ Unchanged (x,y) \<longrightarrow> Unchanged x"
  apply (pref_unlift)
  apply (unfold unch_def before_def after_def nexts_def)
  apply auto 
  done

lemma unch4a: "\<And>w. w \<Turnstile> Unchanged (x,y) \<Longrightarrow> w \<Turnstile> Unchanged x"
proof -
  fix w
  assume a1: "w \<Turnstile> Unchanged (x,y)"
  from a1 unch4[THEN prefD, where w=w and x1=x and y1=y]
  show "w \<Turnstile> Unchanged x" by auto
qed

lemma unch5: "|~ Unchanged (x,y) \<longrightarrow> Unchanged y"
  apply (pref_unlift)
  apply (unfold unch_def before_def after_def nexts_def)
  apply auto 
  done

lemma unch6: "|~ Unchanged (x,y) \<longrightarrow> Unchanged x \<and> Unchanged y"
  apply (pref_unlift)
  apply (unfold unch_def before_def after_def nexts_def)
  apply auto 
  done

lemma unch7:  assumes h1: "|~ Unchanged (x,y)" 
             shows "|~ Unchanged x \<and> Unchanged y"
  using h1
  apply (unfold unch_def before_def after_def nexts_def)
  apply (rule prefI)
  apply auto   
  done

lemma unch8:  assumes h1: "|~ Unchanged x \<and> Unchanged y"
  shows "|~ Unchanged (x,y)" 
  using h1
  apply (unfold unch_def before_def after_def nexts_def)
  apply (rule prefI)
  apply auto   
  done

lemma unch9: "|~ Unchanged (x,y) = (Unchanged x \<and> Unchanged y)"
  apply (pref_unlift)
  apply (unfold unch_def before_def after_def nexts_def)
  apply auto 
  done

lemmas unch_eq1 = unch9[THEN pref_eq]
lemmas unch_eq2 = unch9[THEN prefeq_reflection]

thm unch_eq1
thm unch_eq2


lemma angle_actrans_sem: "|~ (\<langle>F\<rangle>_v) = (F \<and> ((v$) \<noteq> $v))"
  apply (unfold angle_actrans_def actrans_def unch_def)
  apply (pref_unlift)
  apply auto
  done

lemmas angle_actrans_sem_eq = angle_actrans_sem[THEN pref_eq]

lemma "|~ Unchanged (x,y,z) = ($x=x$)"
proof (simp only: unch_eq2)
oops

constdefs

unch1  :: "('a,'b) formfun \<Rightarrow> 'a formula" 
"unch1 v \<equiv> \<lambda> s. s \<Turnstile> v = (nexts v)"

syntax
 "_unch1"     :: "lift \<Rightarrow> lift"  ("(Unchanged1 _)" [100] 99) 

translations
 "_unch1"     \<rightleftharpoons> "unch1" 

section "Lemmas with after"

lemma after_const: "|~ (#c)` = #c"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_fun1: "|~ f<x>` = f <x`>"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_fun2: "|~ f<x,y>` = f <x`,y`>"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_fun3: "|~ f<x,y,z>` = f <x`,y`,z`>"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_fun4: "|~ f<x,y,z,zz>` = f <x`,y`,z`,zz`>"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_forall: "|~ (\<forall> x. P x)` = (\<forall> x. (P x)`)"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_exists: "|~ (\<exists> x. P x)` = (\<exists> x. (P x)`)"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_exists1: "|~ (\<exists>! x. P x)` = (\<exists>! x. (P x)`)"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_pair: "|~ (X,Y)` = (X`,Y`)"
  by (pref_unlift)(auto simp add: before_def nexts_def after_def)

lemma after_le: "|~ (x < y)` = (x` < y`)"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_le_eq: "|~ (x \<le> y)` = (x` \<le> y`)"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_mem: "|~ (x \<in> X)` = (x` \<in> X`)"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_notmem: "|~ (x \<notin> X)` = (x` \<notin> X`)"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_if: "|~ (if T then A else B)` = (if T` then A` else B`)"
  by (pref_unlift)(auto simp add: nexts_def before_def after_def)

lemma after_plus: "|~ (x + y)` = x` + y`"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_minus: "|~ (x - y)` = x` - y`"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_times: "|~ (x * y)` = x` * y`"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_div: "|~ (x div y)` = x` div y`"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_mod: "|~ (x mod y)` = x` mod y`"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_finset: "|~ {X}` = {X`}"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_cons: "|~ (x#xs)` = (x`#xs`)"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_app: "|~ (xs@ys)` = (xs`@ys`)"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemma after_list: "|~ [L]` = [L`]"
  by (pref_unlift) (auto simp add: before_def after_def nexts_def)

lemmas all_after = after_const after_fun1 after_fun2 after_fun3 after_fun4
  after_forall after_exists after_exists1 after_pair after_le
  after_le_eq after_mem after_notmem after_if after_finset after_cons
  after_app after_list

lemmas all_after_unl = all_after[THEN prefD]
lemmas all_after_eq = all_after[THEN prefeq_reflection]

section "Lemmas with before"
 
lemma before_const: "\<turnstile> $(#c) = #c"
  by (int_unlift)(auto simp add: before_def)

lemma before_fun1: "\<turnstile> $(f<x>) = f <$x>"
  by (int_unlift)(auto simp add: before_def)

lemma before_fun2: "\<turnstile> $(f<x,y>) = f <$x,$y>"
  by (int_unlift)(auto simp add: before_def)

lemma before_fun3: "\<turnstile> $(f<x,y,z>) = f <$x,$y,$z>"
  by (int_unlift)(auto simp add: before_def)

lemma before_fun4: "\<turnstile> $(f<x,y,z,zz>) = f <$x,$y,$z,$zz>"
  by (int_unlift)(auto simp add: before_def)

lemma before_forall: "\<turnstile> $(\<forall> x. P x) = (\<forall> x. $(P x))"
  by (int_unlift)(auto simp add: before_def)

lemma before_exists: "\<turnstile> $(\<exists> x. P x) = (\<exists> x. $(P x))"
  by (int_unlift)(auto simp add: before_def)

lemma before_exists1: "\<turnstile> $(\<exists>! x. P x) = (\<exists>! x. $(P x))"
  by (int_unlift)(auto simp add: before_def)

lemma before_pair: "\<turnstile> $(X,Y) = ($X,$Y)"
  by (int_unlift)(auto simp add: before_def)

lemma before_le: "\<turnstile> $(x < y) = ($x < $y)"
  by (int_unlift) (auto simp add: before_def)

lemma before_le_eq: "\<turnstile>  $(x \<le> y) = ($x \<le> $y)"
  by (int_unlift) (auto simp add: before_def)

lemma before_mem: "\<turnstile> $(x \<in> X) = ($x \<in> $X)"
  by (int_unlift) (auto simp add: before_def)

lemma before_notmem: "\<turnstile> $(x \<notin> X) = ($x \<notin> $X)"
  by (int_unlift) (auto simp add: before_def)

lemma before_if: "\<turnstile> $(if T then A else B) = (if $T then $A else $B)"
  by (int_unlift)(auto simp add: before_def)

lemma before_plus: "\<turnstile> $(x + y) = ($x) + ($y)"
  by (int_unlift) (auto simp add: before_def)

lemma before_minus: "\<turnstile> $(x - y) = $x - $y"
  by (int_unlift) (auto simp add: before_def)

lemma before_times: "\<turnstile> $(x * y) = $x * $y"
  by (int_unlift) (auto simp add: before_def)

lemma before_div: "\<turnstile> $(x div y) = $x div $y"
  by (int_unlift) (auto simp add: before_def)

lemma before_mod: "\<turnstile> $(x mod y) = $x mod $y"
  by (int_unlift) (auto simp add: before_def)

lemma before_finset: "\<turnstile> ${X} = {$X}"
  by (int_unlift) (auto simp add: before_def)

lemma before_cons: "\<turnstile>  $(x#xs) = ($x#$xs)"
  by (int_unlift) (auto simp add: before_def)

lemma before_app: "\<turnstile> $(xs@ys) = ($xs@$ys)"
  by (int_unlift) (auto simp add: before_def)

lemma before_list: "\<turnstile> $[L] = [$L]"
  by (int_unlift) (auto simp add: before_def)

lemmas all_before = before_const before_fun1 before_fun2 before_fun3 before_fun4
  before_forall before_exists before_exists1 before_pair before_le
  before_le_eq before_mem before_notmem before_if before_finset before_cons
  before_app before_list

thm all_before

lemmas all_before_unl = all_before[THEN intD]
lemmas all_before_eq = all_before[THEN inteq_reflection]

section "Some general properties"

lemma angle_actrans_def2: "|~ (\<langle>F\<rangle>_v) = (F \<and> v$ \<noteq> $v)"
  by (unfold angle_actrans_def actrans_def unch_def)(pref_unlift,auto)

lemma angle_actrans_conj: "|~ (\<langle>F \<and> G\<rangle>_v) = (\<langle>F\<rangle>_v \<and> \<langle>G\<rangle>_v)"
  by (unfold angle_actrans_def actrans_def unch_def)(pref_unlift,auto)

lemma angle_actrans_disj: "|~ (\<langle>F \<or> G\<rangle>_v) = (\<langle>F\<rangle>_v \<or> \<langle>G\<rangle>_v)"
  by (unfold angle_actrans_def actrans_def unch_def)(pref_unlift,auto)

thm "all_before_eq"

end
